import json
from typing import TYPE_CHECKING, Any, Dict, Generic, Optional, Set, Tuple, Type, TypeVar, cast

import yaml
from pydantic import BaseModel

if TYPE_CHECKING:
    from .manager import Manager

F = TypeVar("F")


class Field(Generic[F]):
    name: str

    def __get__(self, obj: "BalsamModel", type: "Optional[Type[BalsamModel]]" = None) -> F:
        if obj._state == "clean":
            # The field is clean and readable (fetched from REST API):
            if hasattr(obj._read_model, self.name):
                return getattr(obj._read_model, self.name)  # type: ignore
            # The field is write-only but not yet written to:
            elif obj._update_model_cls and self.name in obj._update_model_cls.__fields__:
                return cast(F, None)
            else:
                raise AttributeError(f"Cannot access Field {self.name}")
        elif obj._state == "creating":
            # The field is writeable at object creation time (not yet sent to REST API):
            if hasattr(obj._create_model, self.name):
                return getattr(obj._create_model, self.name)  # type: ignore
            # The field is readable but not yet established by creation:
            elif self.name in obj._read_model_cls.__fields__:
                return cast(F, None)
            else:
                raise AttributeError(f"Cannot access Field {self.name}")
        else:
            # The field has been mutated locally:
            if self.name in obj._dirty_fields:
                return getattr(obj._update_model, self.name)  # type: ignore
            # The field is clean and readable (fetched from REST API):
            elif hasattr(obj._read_model, self.name):
                return getattr(obj._read_model, self.name)  # type: ignore
            # The field is write-only but not yet written to:
            elif obj._update_model_cls and self.name in obj._update_model_cls.__fields__:
                return cast(F, None)
            else:
                raise AttributeError(f"Cannot access Field {self.name}")

    def __set__(self, obj: "BalsamModel", value: F) -> None:
        if obj._state == "creating":
            if self.name not in obj._create_model.__fields__:  # type: ignore
                raise AttributeError(f"Cannot set {self.name} when creating {obj._modelname}")
            setattr(obj._create_model, self.name, value)
        else:
            if obj._update_model_cls is None:
                raise AttributeError(f"{obj.__class__.__name__} is read-only")
            if obj._update_model is None:
                obj._update_model = obj._update_model_cls()
            if self.name not in obj._update_model.__fields__:
                raise AttributeError(f"Cannot set {self.name} when updating {obj._modelname}")
            setattr(obj._update_model, self.name, value)
            obj._dirty_fields.add(self.name)
            obj._state = "dirty"


T = TypeVar("T", bound="BalsamModel")


class BalsamModelMeta(type):
    def __new__(mcls, name: str, bases: Tuple[Any, ...], attrs: Dict[str, Any]) -> "BalsamModelMeta":
        if bases == (object,) or bases == () or "_read_model_cls" not in attrs:
            return super().__new__(mcls, name, bases, attrs)

        field_names = set()
        for model_cls in [
            attrs["_create_model_cls"],
            attrs["_update_model_cls"],
            attrs["_read_model_cls"],
        ]:
            if model_cls is not None:
                field_names.update(
                    {
                        name
                        for name, field in model_cls.__fields__.items()
                        if not field.field_info.extra.get("no_descriptor")
                    }
                )
        for field_name in field_names:
            if field_name in attrs:
                attrs[field_name].name = field_name
        cls = super().__new__(mcls, name, bases, attrs)
        return cls


class BalsamModel(metaclass=BalsamModelMeta):
    _create_model_cls: Optional[Type[BaseModel]]
    _update_model_cls: Optional[Type[BaseModel]]
    _read_model_cls: Type[BaseModel]
    _create_model: Optional[BaseModel]
    _update_model: Optional[BaseModel]
    _read_model: Optional[BaseModel]
    objects: "Manager"  # type: ignore
    id: Any

    def __init__(self, _api_data: bool = False, **kwargs: Any) -> None:
        self._create_model = None
        self._update_model = None
        self._read_model = None
        self._state = None
        self._dirty_fields: Set[str] = set()

        if _api_data:
            self._read_model = self._read_model_cls(**kwargs)
            self._state = "clean"
        else:
            if self._create_model_cls is None:
                raise ValueError(f"{self._modelname} is read only")
            self._create_model = self._create_model_cls(**kwargs)
            self._state = "creating"

    def _set_clean(self) -> None:
        self._state = "clean"
        self._dirty_fields.clear()
        self._update_model = None

    @property
    def _modelname(self) -> str:
        return self.__class__.__name__

    @classmethod
    def _from_api(cls: Type[T], data: Any) -> T:
        return cls(_api_data=True, **data)

    def _refresh_from_dict(self, data: Dict[Any, Any]) -> None:
        self._read_model = self._read_model_cls(**data)
        self._set_clean()

    def save(self) -> None:
        if self._state == "dirty":
            self.__class__.objects._do_update(self)
        elif self._state == "creating":
            assert self._create_model is not None
            created = self.__class__.objects._create(instance=self)
            assert created._read_model is not None
            self._refresh_from_dict(created._read_model.dict())
            self._create_model = None

    def refresh_from_db(self) -> None:
        if self._state == "creating":
            raise AttributeError("Cannot refresh instance before it's saved")
        from_db = self.__class__.objects.all()._get(id=self.id)
        assert from_db._read_model is not None
        self._read_model = from_db._read_model.copy()
        self._set_clean()

    def delete(self) -> None:
        if self._state == "creating":
            raise AttributeError("Cannot delete instance that hasn't been saved")
        self.__class__.objects._do_delete(self)

    class DoesNotExist(Exception):
        def __init__(self, filters: Dict[str, Any]) -> None:
            super().__init__(f"No results matched the query params: {filters}")

    class MultipleObjectsReturned(Exception):
        def __init__(self, nobj: int) -> None:
            super().__init__(f"Returned {nobj} objects; expected one!")

    def display_model(self) -> BaseModel:
        if self._state == "creating":
            assert self._create_model is not None
            return self._create_model
        elif self._state == "dirty":
            assert self._read_model is not None and self._update_model is not None
            return self._read_model.copy(update=self._update_model.dict(exclude_unset=True))
        else:
            assert self._read_model is not None
            return self._read_model

    def display_dict(self) -> Any:
        """Prettify through smart JSON serializer"""
        return json.loads(self.display_model().json())

    def __repr__(self) -> str:
        args = ", ".join(f"{k}={v}" for k, v in self.display_dict().items())
        return f"{self._modelname}({args})"

    def __str__(self) -> str:
        d = self.display_dict()
        return yaml.dump(d, sort_keys=False, indent=4)  # type: ignore

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BalsamModel):
            return False
        return self._state == "clean" and other._state == "clean" and self._read_model == other._read_model


class NonCreatableBalsamModel(BalsamModel):
    id: Field[int]


class CreatableBalsamModel(BalsamModel):
    id: Field[Optional[int]]



from uuid import uuid4

import pytest
from fastapi import status
from fastapi.testclient import TestClient

import balsam.server
from balsam.client import urls
from balsam.server import models
from balsam.server.models.crud import users

from .util import BalsamTestClient


@pytest.fixture(scope="function")
def db_session(setup_database):
    session = next(models.get_session())
    yield session
    session.close()


@pytest.fixture(scope="function")
def fastapi_user_test_client(setup_database, db_session):
    created_users = []

    def _client_factory():
        login_credentials = {"username": f"user{uuid4()}", "password": "test-password"}
        user = users.create_user(db_session, **login_credentials)
        db_session.commit()
        created_users.append(user)

        balsam.server.settings.auth.login_methods = ["password"]
        from balsam.server.main import app

        client = BalsamTestClient(TestClient(app))
        data = client.post_form(urls.PASSWORD_LOGIN, check=status.HTTP_200_OK, **login_credentials)
        token = data["access_token"]
        client.headers.update({"Authorization": f"Bearer {token}"})
        return client

    yield _client_factory
    delete_ids = [user.id for user in created_users]
    db_session.query(models.User).filter(models.User.id.in_(delete_ids)).delete(synchronize_session=False)
    db_session.commit()


@pytest.fixture(scope="function")
def anon_client(setup_database):
    from balsam.server.main import app

    return BalsamTestClient(TestClient(app))


@pytest.fixture(scope="function")
def auth_client(fastapi_user_test_client):
    return fastapi_user_test_client()



import os
from typing import List, Optional

from .compute_node import ComputeNode


class CooleyNode(ComputeNode):

    cpu_ids = list(range(12))
    gpu_ids = list(range(2))

    @classmethod
    def get_job_nodelist(cls) -> List["CooleyNode"]:
        """
        Get all compute nodes allocated in the current job context
        """
        nodefile = os.environ["COBALT_NODEFILE"]
        # a file containing a list of node hostnames, one per line
        with open(nodefile) as fp:
            data = fp.read()
        splitter = "," if "," in data else None
        hostnames = data.split(splitter)
        hostnames = [h.strip().split(".")[0] for h in hostnames if h.strip()]
        return [cls(hostname, hostname) for hostname in hostnames]

    @staticmethod
    def get_scheduler_id() -> Optional[int]:
        id = os.environ.get("COBALT_JOBID")
        if id is not None:
            return int(id)
        return None



import logging
import secrets
from typing import Any, Dict, Optional
from urllib.parse import urlencode

import requests
from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.responses import RedirectResponse
from sqlalchemy.orm import Session, exc

from balsam.schemas import UserOut
from balsam.server import settings
from balsam.server.models import get_session
from balsam.server.models.crud import users

logger = logging.getLogger(__name__)


router = APIRouter(prefix="/oauth")


def generate_state(user_code: Optional[str] = None) -> str:
    """
    Produce new random state to store in DB. If Device Code flow, we have a
    user_code and the state is associated with a user_code.
    """
    secret = secrets.token_urlsafe()
    if user_code is None:
        return f"BROWSER-FLOW {secret}"
    return f"DEVICE-FLOW {user_code} {secret}"


def state_to_usercode(state: str) -> Optional[str]:
    """
    Extract user_code from DB state.  Returns None
    if this is not part of a Device Code flow.
    """
    if state.startswith("DEVICE-FLOW"):
        return state.split(" ")[1]
    return None


def alcf_username_from_token(token: str) -> str:
    """
    Trade ALCF-OAuth access token for user info.
    Returns the username for Balsam to get a unique identity.
    """
    conf = settings.auth.oauth_provider
    assert conf is not None
    resp = requests.get(
        conf.user_info_uri,
        headers={"Authorization": f"Bearer {token}"},
    )
    logger.debug(f"user info response: {resp.status_code}\n{resp.text}")
    dat = resp.json()
    logger.debug(f"user info response json: {dat}")
    return str(dat["username"])


def redirect_to_oauth_provider(request: Request, state: str) -> RedirectResponse:
    """
    Direct the user agent to the OAuth provider to authorize.
    Used by both pure-Authorization code and hybrid Device Code flows.
    """
    conf = settings.auth.oauth_provider
    assert conf is not None, "Using OAuth without configuration"
    callback_uri = f"{conf.redirect_scheme}://{request.url.netloc}" + conf.redirect_path
    params = urlencode(
        dict(
            response_type="code",
            client_id=conf.client_id,
            redirect_uri=callback_uri,
            scope=conf.scope,
            state=state,
        )
    )
    return RedirectResponse(conf.request_uri + "?" + params)


@router.get("/ALCF/login/device")
def start_login_device(request: Request, user_code: str, db: Session = Depends(get_session)) -> RedirectResponse:
    """
    User provides user code to proceed with "device code flow" (started by a CLI login client)
    Balsam hands off to ALCF OAuth for authenticating user via "authorization code flow"
    """
    try:
        device_code_attempt = users.get_device_code_attempt_by_user(db, user_code)
    except exc.NoResultFound:
        raise HTTPException(status_code=400, detail="invalid_client")
    state = generate_state(user_code=device_code_attempt.user_code)
    users.add_auth_state(db, state)
    db.commit()
    return redirect_to_oauth_provider(request, state)


@router.get("/ALCF/login")
def start_login(request: Request, db: Session = Depends(get_session)) -> RedirectResponse:
    """
    User initiated Web browser authorization code flow
    """
    state = generate_state(user_code=None)
    users.add_auth_state(db, state)
    db.commit()
    return redirect_to_oauth_provider(request, state)


@router.get("/ALCF/callback")
def callback(
    request: Request,
    code: str,
    state: str,
    error: Optional[str] = None,
    error_description: Optional[str] = None,
    db: Session = Depends(get_session),
) -> Dict[str, Any]:
    """
    https://tools.ietf.org/html/rfc6749#section-4.1.2.1
    Callback from OAuth provider.  This is used by both pure-Authorization code
    and hybrid Device code flows.
    """
    # If user_code is not None, we're on the second half of a device-code login flow
    user_code = state_to_usercode(state)

    if error == "access_denied" and user_code is not None:
        users.deny_device_code_attempt(db, user_code)
        db.commit()

    if error:
        logger.warning(f"OAuth error response: {error}")
        if error_description:
            logger.warning(f"OAuth error description: {error_description}")
        return {"error": error, "error_description": error_description}

    # Ensure the "state" is in our DB to mitigate CSRF
    try:
        users.verify_auth_state(db, state)
    except exc.NoResultFound:
        raise HTTPException(status_code=400, detail="invalid_state")

    # Use `code` to request an access token from the OAuth provider
    conf = settings.auth.oauth_provider
    assert conf is not None, "Using OAuth without configuration"
    callback_uri = f"{conf.redirect_scheme}://{request.url.netloc}" + conf.redirect_path
    token_params = dict(
        client_id=conf.client_id,
        client_secret=conf.client_secret,
        code=code,
        redirect_uri=callback_uri,
        grant_type="authorization_code",
    )
    logger.debug(f"Requesting token from {conf.token_uri}: {token_params}")
    token_response = requests.post(
        conf.token_uri,
        data=token_params,
        headers={"Cache-Control": "no-cache"},
    )
    logger.debug(f"Received token response: {token_response.status_code}")
    token_data = token_response.json()
    logger.debug(f"token response json: {token_data}")
    access_token = str(token_data["access_token"])

    # Use access token to get ALCF username
    username = alcf_username_from_token(access_token)
    try:
        from_db = users.get_user_by_username(db, username)
        user = UserOut(id=from_db.id, username=from_db.username)
        logger.info(f"Looked up existing username: {username}")
    except exc.NoResultFound:
        user = users.create_user(db, username, password=None)
        db.commit()
        logger.info(f"Created new Balsam username: {username}")

    # If this is coupled to a Device Code login flow, authorize the device code
    # attempt
    if user_code:
        logger.info(f"Authorizing device code attempt for user code: {user_code}")
        users.authorize_device_code_attempt(db, user_code, user)
        db.commit()
    else:
        logger.debug("No user code; this is an ordinary authorization code callback.")
    return {"username": username}
